package asdraw;

import geometry.*;
import colors.*;

import java.awt.Insets;
import java.awt.event.*;
import java.awt.Color;

import javax.swing.JButton;
import javax.swing.Timer;
import javax.swing.WindowConstants;

import java.lang.reflect.Array;
import java.util.*;

import tunes.*;


import asdraw.MyKeyAdapter;
import asdraw.MyTimer;

/**
 * Copyright 2009 Viera K. Proulx
 * This program is distributed under the terms of the 
 * GNU Lesser General Public License (LGPL)
 */


/**
 * World for programming interactive games - with graphics, key events,
 * timer, and sound effects. 
 * 
 * @author Viera K. Proulx
 * @since October 29, 2009
 */
abstract public class World implements SoundConstants{

  /** the canvas that displays the current world */
  public Canvas theCanvas;

  /** true if 'bigBang' started the world and it did not end, did not stop */
  protected boolean worldExists = false;

  /** true if 'bigBang' started the world and the applet stopped the world */
  protected boolean worldStopped = false;

  /** the timer for this world */
  protected MyTimer mytime;
  
  /** timer events not processed when the mouse event is processed */
  protected boolean stopTimer = false;
  
  /** the button to set to CLEAR when the world ends */
  protected JButton bigButton;
  
  /** the key adapter for this world*/
  private MyKeyAdapter ka;
  
  /** the mouse adapter for this world */
  private MyMouseAdapter ma;
  
  /** the window closing listener for this world */
  private WindowListener windowClosing;
  
  /////////////////////////////////////////////////////////////////////////
  // Data for handling the sound                                         //
  /////////////////////////////////////////////////////////////////////////
  
  /** the MIDI synthesizer that plays the notes */
  private MusicBox myMusicBox;

  /** the collection of tunes to play on tick */
  public TuneBucket tickTunes;
  
  /** the collection of tunes currently playing on tick */
  protected TuneBucket currentTickTunes;
  
  /** the collection of tunes to play on key event */
  public TuneBucket keyTunes;  
  
  /** the collection of tunes currently playing on key event */
  protected HashMap<String,TuneBucket> keyReleasedTunes;
  
  /**
   * The default constructor. To start the world one must invoke the
   * <code>bigBang</code> method.
   */
  public World(){ 
    this.initMusic();
  }

  /////////////////////////////////////////////////////////////////////////
  // Methods for interacting with the World                              //
  /////////////////////////////////////////////////////////////////////////
  /**
   * Initialize the world with the canvas and the music synthesizer
   */
  protected void initWorld(Canvas theCanvas, 
                        JButton bigButton){
    this.theCanvas = theCanvas;
    this.bigButton = bigButton;
    //this.initMusic();   
  }
  
  /**
   * Initialize the MIDI synthesizer and the tune buckets
   */
  protected void initMusic(){
    /** the MIDI synthesizer that plays the notes */
    this.myMusicBox = new MusicBox();

    if (tunes.MusicBox.SYNTH_READY){
      /** the collection of tunes to play on tick */
      this.tickTunes = new TuneBucket(this.myMusicBox);

      /** the collection of tunes currently playing on tick */
      this.currentTickTunes = new TuneBucket(this.myMusicBox);

      /** the collection of tunes to play on key event */
      this.keyTunes= new TuneBucket(this.myMusicBox);  

      /** the collection of tunes currently playing on key event */
      this.keyReleasedTunes = new HashMap<String, TuneBucket>();
    }
    else{
      /** notify the user that music cannot play */
      System.out.println("MIDI synthesizer or the soundbank not available.");
      System.out.println("Tunes will not be played.");
      
    }
  }
  
  /**
   * <p>Change the program for the {link MusicBox MusicBox} to the given
   * list of instruments.</p>
   * <p>If there are fewer than 16 instruments given, all the remaining ones
   * are set to play piano.</p>
   * 
   * @param myInstruments an <code>ArrayList</code> of 16 instrument numbers, 
   * must be in the range from 1 to 128
   */
  public void programChange(ArrayList<Integer> myInstruments){
    if (tunes.MusicBox.SYNTH_READY){
      int[] instruments = new int[16];
      if (myInstruments.size() < 16){
        for (int i = myInstruments.size(); i < 16; i++){
          myInstruments.add(1);
        }
      }
      for (int i = 0; i < 16; i++){
        instruments[i] = myInstruments.get(i);
      }
      this.myMusicBox.initChannels(instruments);
    }
  }
  /**
   * <p>Change the program for the {link MusicBox MusicBox} to the given
   * list of instruments.</p>
   * <p>If there are fewer than 16 instruments given, all the remaining ones
   * are set to play piano.</p>
   * 
   * @param myInstruments an <code>Array</code> of 16 instrument numbers, 
   * must be in the range from 1 to 128
   */
  public void programChange(int[] myInstruments){
    if (tunes.MusicBox.SYNTH_READY){
      // make sure the program change represents 16 channels
      int[] okInstruments = new int[16];
      for (int i = 0; i < Math.min(16, Array.getLength(myInstruments)); i++){
        okInstruments[i] = myInstruments[i];
      }
      if (Array.getLength(myInstruments) < 16){
        for (int i = Array.getLength(myInstruments); i < 16; i++){
          okInstruments[i] = 1;
        }
      }
      this.myMusicBox.initChannels(okInstruments);
    }
  }
  
  /** 
   * Pause at the start when building GUIs and installing listeners
   * and at the end so the last tunes are heard
   */
  protected void sleepSome(int miliseconds){
    // pause a bit so that two canvases do not compete when being opened
    // almost at the same time
    long start = System.currentTimeMillis();
    long tmp = System.currentTimeMillis();
    //System.out.println("Going to sleep.");

    while(tmp - start < miliseconds){
      tmp = System.currentTimeMillis();
    }
  }
  
  /**
   * Start the world by creating a canvas of the given size, creating
   * and adding the key and mouse adapters, and starting the timer at the
   * given speed.
   * 
   * @param w the width of the <code>{@link Canvas Canvas}</code>
   * @param h the height of the <code>{@link Canvas Canvas}</code>
   * @param speed the speed at which the clock runs
   */
  protected void bigBang(int w, int h, double speed){

    this.worldExists = true;
    
    //this.initMusic();
    

    // set the drawing size to the given width and height
    this.theCanvas.painter.setSize(w, h);
    
    
    // pause a bit so that two canvases do not compete when being opened
    // almost at the same time
    this.sleepSome(100); // miliseconds
    
    // add the key listener to the frame for our canvas
    this.ka = new MyKeyAdapter(this);
    this.theCanvas.painter.addKeyListener(this.ka);
    
    // add the mouse listener to the frame for our canvas
    this.ma = new MyMouseAdapter(this);
    this.theCanvas.painter.addMouseListener(this.ma);

    // make sure the canvas can respond to the key events
    this.theCanvas.setFocusable(true);
   
    // pause again after the Canvas is shown to make sure 
    // all listeners and the timer are installed for theCanvas
    this.sleepSome(100); //miliseconds

    // add the timer
    this.mytime = new MyTimer(this, speed);

    // start it if speed is greater than 0
    if (speed > 0.0)
      this.mytime.timer.start();

    // finally, show the canvas and draw the initial world
    this.theCanvas.addNotify(); 
    this.theCanvas.setVisible(true);
    
    this.tickTunes.playTunes();
    this.keyTunes.playTunes();
    
    this.drawWorld();
    
    // note that the applet has not been stopped while the world runs
    this.worldStopped = false;
  }


  /**
   * Stop the timer, disable event processing - leave the canvas open, 
   * write the end of time message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   */
  public void endOfTime(String s){
    this.endOfTime(s, Color.black);
  }

  /**
   * Stop the timer, disable event processing - leave the canvas open, 
   * write the end of time message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   * @param col the <CODE>IColor</CODE> for the string
   */
  public void endOfTime(String s, IColor col){
    this.endOfTime(s, col.thisColor());
  }
  
  /**
   * Stop the timer, disable event processing - leave the canvas open, 
   * write the end of time message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   * @param col the <CODE>IColor</CODE> for the string
   */
  public void endOfTime(String s, Color col){
    if (this.worldExists){
      // remove listeners and set worldExists to false
      this.mytime.timer.stop();
      this.worldExists = false;
      this.worldStopped = true;
      this.mytime.stopTimer();
      this.theCanvas.painter.removeKeyListener(this.ka);
    

      System.out.println("The world stopped: " + s);

      if (this.theCanvas != null){
        // draw the final scene of the world with the end of time message
        this.theCanvas.clear();
        this.draw();
        this.theCanvas.drawString(new Posn(10, 20),s, col);
      }
      if (tunes.MusicBox.SYNTH_READY){ 
        // play the final drumroll
        this.tickTunes.playTunes();

        this.sleepSome(1000); // miliseconds

        // stop the playing of tunes in the tickTune bucket and the keyTunes bucket
        this.tickTunes.clearBucket();
        this.keyTunes.clearBucket();
      }
      
      // 'hit the STOP button' in the applet controls
      this.bigButton.setText(WorldApplet.CLEAR);
    }
  }


  /**
   * End the world interactions - leave the canvas open, 
   * write the end of world message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   * @param col the <CODE>IColor</CODE> for the string
   */
  public void endOfWorld(String s, IColor col){
    this.endOfWorld(s, col.thisColor());
  }

  /**
   * End the world interactions - leave the canvas open, 
   * write the end of world message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   */
  public void endOfWorld(String s){
    this.endOfWorld(s, Color.black);
  }
  /**
   * End the world interactions - leave the canvas open, 
   * write the end of world message onto the canvas
   * Play the final tunes in the tickTunes bucket
   * 
   * @param s the string to display at the end
   * @param col the <CODE>Color</CODE> for the string
   */
  public void endOfWorld(String s, Color col){
    if (this.worldExists){
      // remove listeners and set worldExists to false
      this.mytime.timer.stop();
      this.mytime.stopTimer();
      this.worldExists = false;
      this.theCanvas.painter.removeKeyListener(this.ka);
    }
    
    System.out.println("The end of the world: " + s);
    
    if (this.theCanvas != null){
      // draw the final scene of the world with the end of time message
      this.theCanvas.clear();
      this.draw();
      this.theCanvas.drawString(new Posn(10, 20),s, col);
    }
    if (tunes.MusicBox.SYNTH_READY){ 
      // play the final drumroll
      this.tickTunes.playTunes();

      this.sleepSome(1000); // miliseconds

      // stop the playing of tunes in the tickTune bucket and the keyTunes bucket
      this.tickTunes.clearBucket();
      this.keyTunes.clearBucket();

      
      // 'hit the STOP button' in the applet controls
      this.bigButton.setText(WorldApplet.CLEAR);
    }
  }

  /**
   * The method invoked by the timer on each tick.
   * Delegates to the user to define a new state of the world,
   * then resets the canvas and event handlers for the new world
   * to those currently used.
   */
  protected void processTick(){
    try{
      if (this.worldExists && !this.stopTimer){

        if (tunes.MusicBox.SYNTH_READY){ 
          // advance the tick on current tunes
          // and stop playing those that are done
          this.currentTickTunes.nextBeat();
          this.tickTunes.clearBucket();
        }
        
        // process the changes to the world on this tick
        this.onTick();

        if (tunes.MusicBox.SYNTH_READY){ 
          // play the tunes collected in the tick tune bucket
          this.tickTunes.playTunes();
          this.currentTickTunes.add(this.tickTunes);
        }
        // draw the world as it looks after this tick
        this.drawWorld();
      }
      // the world has stopped - clear the tick tune and key tunes buckets
      else {
        if (tunes.MusicBox.SYNTH_READY){ 
          // clear the tick tune bucket
          this.tickTunes.clearBucket();
          // clear the key tune bucket
          this.keyTunes.clearBucket();
          }
      }
  } 
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }  

  /**
   * <P>User defined method to be invoked by the timer on each key event.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   */
  abstract public void onTick();
    

  /**
   * The method invoked by the key adapter on key pressed event.
   * Clears the contents of the keyTunes bucket.
   * Delegates to the user to define a new state of the world,
   * then plays the tunes deposited in the keyTunes bucket.
   * It then copies these tunes to the keyReleasedTunes hashmap,
   * using the key pressed as the hashmap key.
   */
  protected void processKeyEvent(String ke){
    try{
      if (this.worldExists){

        // empty the key tune bucket
        this.keyTunes.clearTunes();

        // process the changes to the world on this key event
        this.onKeyEvent(ke);

        // play the tunes collected in the key tune bucket
        // save what is currently playing so it plays for the desired
        // number of ticks
        this.keyReleasedTunes.put(ke, this.keyTunes.copy());
        this.keyTunes.playTunes();
        
        // draw the world as it looks after this tick
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * The method invoked by the key adapter when key is released.
   * Delegates to the user to define a new state of the world,
   * then removes from the KeyReleasedTunes bucket those tunes that 
   * started playing when the given key was pressed.
   * 
   */
  void processKeyReleasedEvent(String ke){
    TuneBucket tmp = this.keyReleasedTunes.remove(ke);
    if (tmp != null){
      tmp.clearBucket();
    }
    // invoke user-defined onKeyReleased method
    this.onKeyReleased(ke);
  }

  /**
   * <P>User defined method to be invoked by the key adapter 
   * when a key is released.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   */  
  public void onKeyReleased(String ke){}

  /**
   * <P>User defined method to be invoked by the key adapter 
   * when a key is pressed.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Implement this method in the game world class</P>
   */
  abstract public void onKeyEvent(String s);
 
  /**
   * The method invoked by the mouse adapter on mouse clicked event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when clicked
   */
  protected void processMouseClicked(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse clicked event
        this.onMouseClicked(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is clicked.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when clicked
   */
  public void onMouseClicked(Posn mouse){}
  
  /**
   * The method invoked by the mouse adapter on mouse entered event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when entered
   */
  protected void processMouseEntered(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse entered event
        this.onMouseEntered(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is entered.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when entered
   */
  public void onMouseEntered(Posn mouse){}  
  
  
  /**
   * The method invoked by the mouse adapter on mouse exited event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when exited
   */
  protected void processMouseExited(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse exited event
        this.onMouseExited(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is exited.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when exited
   */
  public void onMouseExited(Posn mouse){}  
  
  
  /**
   * The method invoked by the mouse adapter on mouse pressed event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when pressed
   */
  protected void processMousePressed(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse pressed event
        this.onMousePressed(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is pressed.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when pressed
   */
  public void onMousePressed(Posn mouse){}
  
  /**
   * The method invoked by the mouse adapter on mouse released event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when released
   */
  protected void processMouseReleased(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse released event
        this.onMouseReleased(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }
  
  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is released.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when released
   */
  public void onMouseReleased(Posn mouse){}

  /**
   * Invoke the user defined <code>draw</code> method, if this 
   * <code>{@link World World}</code> has been initialized 
   * via <code>bigBang</code> and did not stop or end 
   * via <code>endOfTime</code> 
   * or <code>endOfWorld</code>
   */
  protected void drawWorld(){
    if (this.worldExists){
      this.theCanvas.clear();
      this.draw();
      if (!tunes.MusicBox.SYNTH_READY)
        this.theCanvas.drawString(new Posn(20, 20), 
            "MIDI synthesizer or the soundbank not available.");
      
    }
  }
  
  /**
   * <P>User defined method to draw the <code>{@link World World}</code>.</P>
   * <P>Implement this method in the game world class</P>
   */
  abstract public void draw();
  

  /**
   * Stop the world - temporarily or for good:
   * stop the timer, remove key listener, mark the world stopped.
   */
  protected void stopWorld(){
    this.worldExists = false;
    this.worldStopped = true;
    this.mytime.timer.stop();
    this.keyTunes.clearBucket();
    this.tickTunes.clearBucket();
    
    // discard the old key listener
    this.theCanvas.painter.removeKeyListener(this.ka);
  }
}

/**
 * The window closing listener - it stops the timer when the frame is closed
 * and stops playing all tunes - in the tickTunes bucket and in the 
 * keyTunes bucket.
 * 
 * @author Viera K. Proulx
 * @since 5 August 2010
 *
 */
class MyWindowClosingListener extends WindowAdapter{
  World w;
  MyWindowClosingListener(World w){
    this.w = w;
  }
  public void windowClosing(WindowEvent we){
    this.w.mytime.stopTimer();

    if (tunes.MusicBox.SYNTH_READY){ 
      // clear the tick tune bucket
      this.w.tickTunes.clearBucket();
      // clear the key tune bucket
      this.w.keyTunes.clearBucket();
    }
  }
}

/**
 * The action listener for the timer events.
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyTimer{

  /** the current <code>{@link World World}</code> 
   * that handles the timer events */
  protected World currentWorld;

  /** the <code>Timer</code> that generates the time events */
  protected Timer timer;
  
  public boolean running = true;

  /** the timer speed */
  protected int speed;

  /**
   * Create the initial timer for the given 
   * <code>{@link World World}</code> at the given <code>speed</code>.
   * 
   * @param currentWorld the given <code>{@link World World}</code>
   * @param speed the given <code>speed</code>
   */
  protected MyTimer(World currentWorld, double speed){
    this.currentWorld = currentWorld;
    this.timer = new Timer((new Double(speed * 1000)).intValue(), 
        this.timerTasks);
    this.speed = (new Double(speed * 1000)).intValue();
  }

  /**
   * The callback for the timer events
   */
  protected ActionListener timerTasks = new ActionListener() {
    public void actionPerformed(ActionEvent evt) {
      if (running)
        currentWorld.processTick();
    }
  };

  /**
   * A helper method to convert the <code>speed</code> given as 
   * a delay time into milliseconds
   */
  protected void setSpeed(){ 
    this.timer.setDelay(this.speed);
  }
  
  protected void stopTimer(){
    this.running = false;
  }
}


/**
 * <p>The implementation of callbacks for the mouse events: 
 * mouse clicked, entered, exited, pressed, and released.</p>
 * <p>We do not handle mouse motion events: mouse dragged and 
 * mouse moved.</p>
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyMouseAdapter extends MouseAdapter {

  /** the current <code>{@link World World}</code> 
   * that handles the mouse events */
  protected World currentWorld;

  /** the mouse position recorded by the mouse event handler */
  protected Posn mousePosn;
  
  /**
   * Create the mouse listener for the given <code>{@link World World}</code>.
   * 
   * @param currentWorld the given <code>{@link World World}</code>
   */
  protected MyMouseAdapter(World currentWorld){
    this.currentWorld = currentWorld;
  }
  
  // ---------------------------------------------------------------------
  // the mouse event handlers
  
  /**
   * Adjust the reported mouse position to account for the top bar
   * 
   * @param mousePosn the recorded mouse position
   * @return the actual mouse position
   */ 
  Posn adjustMousePosn(Posn mousePosn){
    // .... use this to find the height of the top bar
    Insets ins = this.currentWorld.theCanvas.painter.getInsets();
    mousePosn.y -= ins.top;
    return mousePosn;
  }
   
  /**
   * Invoked when the mouse has been clicked on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseClicked(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseClicked(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse enters a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseEntered(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseEntered(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse exits a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseExited(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseExited(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse button has been pressed on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mousePressed(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMousePressed(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when a mouse button has been released on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseReleased(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseReleased(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
   }
}
