package tunes;

import java.util.*;

/**
 * Copyright 2010 Viera K. Proulx
 * This program is distributed under the terms of the 
 * GNU Lesser General Public License (LGPL)
 */

/**
 * <p>A class to represent a collection of notes to be played
 * on one instrument at the same time.</p>
 * <p>Individual notes do not need to have the same duration.</p>
 * 
 * @author Viera K. Proulx
 * @since 8 June 2010
 */
public class Chord{
  
  /** the list of notes in this chord */
  public ArrayList<Note> notes = new ArrayList<Note>();
  
  /**
   * The default constructor
   */
  public Chord(){}
  
  /**
   * A constructor that creates a <code>Chord</code> from 
   * a sequence of <code>String</code>s that represent
   * notes to include in this chord.
   */
  public Chord(String... snotes){
    for (String s: snotes){
      this.addNote(s);
    }
  }
  
  /**
   * A constructor that creates a <code>Chord</code> from 
   * a sequence of <code>int</code>s that represent
   * notes to include in this chord.
   */
  public Chord(int... pitches){
    for (int pitch: pitches){
      if (pitch == 0)
        this.addNote(pitch, 0);
      else
        this.addNote(pitch, 1);
    }
  }
  
  /**
   * A constructor that creates a <code>Chord</code> from 
   * a sequence of <code>Note</code>s to include in this chord.
   */
  public Chord(Note... notes){
    for (Note n: notes){
      this.addNote(n);
    }
  }
  
  /**
   * Make a deep copy of this chord, so we can mutate the note durations as 
   * the chord plays out.
   * 
   * @return a copy of this chord
   */
  public Chord copy(){
    Chord replica = new Chord();
    for (Note n: this.notes){
      replica.notes.add(new Note(n.snote));
    }
    return replica;
  }
  /** add a note to this chord */
  public void addNote(Note n){
    this.notes.add(n);
  }
  
  /** add a note to this chord, by converting its String representation*/
  public void addNote(String snote){
    this.notes.add(new Note(snote));
  }
  
  /** add a note to this chord, by converting its 
    * pitch and duration representation*/
  public void addNote(int pitch, int duration){
    this.notes.add(new Note(pitch, duration));
  }
  
  /**
   * EFFECT: reduce the duration of all notes in this chord by one.
   * 
   * Produce the chord that represents the list of notes
   * that have stopped playing at the end of this beat.
   * 
   * @return Chord of notes that should stop playing
   */
  public Chord nextBeat(){
    Chord ch = new Chord();
    for (Note n: this.notes){
      n.nextBeat();
      if (n.duration <= 0)
        ch.addNote(n);
    }
    return ch;
  }
  
  /**
   * <p>EFFECT: increase the duration of all notes in this chord by one.</p>
   * 
   * <p>Produce this modified chord.</p>
   * 
   * @return this chord after the note durations have been increased
   */
  public Chord skipBeat(){
    for (Note n: this.notes){
      n.skipBeat();
    }
    return this;
  }
  
  /**
   * Have all notes in this chord finish playing?
   * @return true if there are no notes, or all have duration 0 or less
   */
  public boolean isSilent(){
    for (Note n: this.notes){
      if (n.duration > 0)
        return false;
    }
    return true;
  }
  
  /**
   * Make a new copy of this chord - with new notes
   * 
   * @return new chord with the same note values, but distinct instances
   */
  public Chord copyChord(){
	  Chord result = new Chord();
	  result.notes.clear();
	  for (Note n: this.notes){
		  result.addNote(n);
	  }
	  return result;
  }
  
  
  /** a singleton chord representing silence */
  public static Chord noplay = new Chord();
  
  /** number of notes in this chord */
  public int size(){
    return this.notes.size();
  }
  
  /** does this chord contain the given note?
   * @param n the given note
   * @return true if this chord contains the given note
   */
  public boolean containsNote(Note n){
    for (Note note: this.notes){
      if (note.sameNote(n))
      return true;
    }
    return false;
  }
}